import asyncio
from typing import Callable, Iterable

from ..Log import Log

class Event:
    def __init__(self, funcs: list[Callable[[], None]] = [], error_ignored: bool = False):
        self._funcs: dict[str, Callable[[any], None]] = {}
        if isinstance(funcs, Iterable):
            for func in funcs:
                self.add(func)
        else:
            self.add(funcs)
            
        self._error_ignored = error_ignored
        
    def add(self, func: Callable[[], None]):
        self._funcs[f'{func.__module__}.{func.__name__}_{tuple(func.__annotations__.keys())}'] = func
        
    def invoke(self, *args, **kwargs):
        for func in self._funcs.values():
            try:
                res = func(*args, **kwargs)
                if asyncio.iscoroutine(res):
                    raise 
                
            except Exception as ex:
                if not self._error_ignored:
                    raise ex
                Log.writeError()
                
                
    async def invokeAsync(self, *args, **kwargs):
        for func in self._funcs.values():
            try:
                res = func(*args, **kwargs)
                if asyncio.iscoroutine(res):
                    await res
                    
            except Exception as ex:
                if not self._error_ignored:
                    raise ex
                Log.writeError()


class EventKwargs(Event):
    def add(self, func: Callable[[any], None]):
        super().add(func)
    
    def invoke(self, **kwargs):
        super().invoke(**kwargs)
                
    async def invokeAsync(self, **kwargs):
        super().invokeAsync(**kwargs)


from typing import Union, Iterable
import numpy as np 

from ..Classes import Buffer, Event, Vec2, Vec3
from .Pixel import Pixel
from .Widget import Widget
from .Drawer import Drawer
from .. import Converter
from .. import Func

class Window:    
    def __init__(
            self, 
            size: Union[Vec2[int], Iterable[int]] = None,
            offset_pos: Union[Vec3[int], Iterable[int]] = None, 
            widgets: Union[Iterable[Widget], Widget] = None, 
            clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
            frame: bool = False,
            name: Union[str, None] = None,
            auto_size: bool = False,
            *args, **kwargs
        ):
        ''' size and pos '''
        self._size = Converter.toVec2(size)
        self._auto_size = auto_size
        self._indent_size = Vec2(2, 2) if frame else Vec2(0, 0)
        self._offset_pos = Converter.toVec3(offset_pos)
        
        ''' widgets '''
        self._offset_pos_widgets: Vec3[int] = Vec3(0, 0, 0)
        self._widgets: list[Widget] = []
        for widget in Converter.toListWidgets(widgets):
            self.addWidget(widget)
        # if self._auto_size:
        #     for widget in self._widgets:
        #         widget.resize_event.add(self.setUpdateAutoSize)
        
        ''' pixels '''
        self._clear_pixel = Converter.toPixel(clear_pixel, Pixel())
        
        ''' events '''
        self._open_event: Event = Event()
        self._close_event: Event = Event()
        self._resize_event: Event = Event()
        
        ''' flags '''
        self._flag_updateAutoSize = self._auto_size
        
        ''' other '''
        self._frame: bool = frame
        self._name = name
        
    
    def addWidget(self, widget: Widget):
        self._widgets.append(widget)
        if self._auto_size:
            widget.resize_event.add(self.setUpdateAutoSize)
    
    def render(self) -> Buffer[Pixel]:
        widgets: list[tuple[any]] = []
        for widget in sorted(self._widgets, key=lambda widget: widget.offset_pos.z):
            widgets.append((
                ((widget.offset_pos + (
                    Vec3(1, 1, 0) if self._frame else Vec3(0, 0, 0)
                ) + self._offset_pos_widgets).toTuple()[:2]),
                widget.render()
            ))
        
        if self._flag_updateAutoSize:
            self._updateAutoSize()
        
        if self.size == (0, 0):
            return Buffer.empty
        
        buffer = Buffer(*self.size, self._clear_pixel)
        
        for widget in widgets:
            buffer.paste(*widget[0], widget[1])
        
        if self._frame:
            buffer.paste(
                0, 0, 
                Drawer.frame(
                    *self.size, 
                    self._clear_pixel.front_color,
                    self._clear_pixel.back_color
                ) if self._clear_pixel is not None else
                Drawer.frame(
                    *self.size, 
                    None, None
                ),
                Drawer.mergeFramePixels
            )     
                   
        return buffer
    
    def setUpdateAutoSize(self):
        self._flag_updateAutoSize = True
    
    @property
    def offset_pos(self) -> Vec3[int]:
        return self._offset_pos
    @offset_pos.setter
    def offset_pos(self, value: Vec3[int]):
        self._offset_pos = value
    @property
    def offset_x(self) -> int:
        return self._offset_pos.x
    @property
    def offset_y(self) -> int:
        return self._offset_pos.y
    @property
    def offset_z(self) -> int:
        return self._offset_pos.z
    @property
    def size(self) -> Vec2[int]:
        return self._size
    @size.setter
    def size(self, value: Vec2[int]):
        self._size = value
        self._size.change_event.add(lambda: self._resize_event.invoke())
        self._resize_event.invoke()
    @property
    def width(self) -> int:
        return self.size.x
    @width.setter
    def width(self, value: int):
        self.size.x = value
    @property
    def height(self) -> int:
        return self.size.y
    @height.setter
    def height(self, value: int):
        self.size.y = value
    @property
    def name(self) -> Union[str, None]:
        return self._name
    @property
    def open_event(self) -> Event:
        return self._open_event
    @property
    def close_event(self) -> Event:
        return self._close_event
    @property
    def resize_event(self) -> Event:
        return self._resize_event
    
    def _updateAutoSize(self):
        if len(self._widgets) == 0 or self._auto_size is False:
            if self._auto_size:
                self.size = Vec2(0, 0)
            return
        
        self.size = Func.calculateSizeByItems(self._widgets) + self._indent_size + self._offset_pos_widgets.toTuple()[:2]
        
        self._flag_updateAutoSize = False
    
    
    def __str__(self, **kwargs):
        kwargs.update({
            "name": self._name,
            "size": self._size,
            "offset_pos": self._offset_pos
        })
        return f'{self.__class__.__name__}({'; '.join([f'{key}:{value}' for key, value in kwargs.items()])})'


import asyncio

class GVars:
    ASYNC_EVENT_LOOP = asyncio.get_event_loop()
    
    SPS = 30
    FPS = 15
    
    CLEAR_LINES = 30
        
    DRAWER_MAX_SIZE_CACHE = 5
    
    WRITE_WARNING_DYNAMIC_MODULE = True
    
    DEBUG_SHOW_INPUT_KEY = False
    DEBUG_SHOW_CANCELLED_THREAD_MESSAGE = False
    DEBUG_SHOW_ANSIICOLOR_CHARS = False
    



from typing import Union, Iterable

from ..Window import Window
from ..Widget import Widget
from ..Pixel import Pixel
from ...Classes import Vec3, Vec2, Buffer, Anchor
from ... import Func
from ... import Converter
from ..Drawer import Drawer


class TitledWindow(Window):
    def __init__(
            self,
            size: Union[Vec2[int], Iterable[int]] = None, 
            offset_pos: Union[Vec3[int], Iterable[int]] = None, 
            widgets: Union[Iterable[Widget], Widget] = None, 
            clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
            frame: bool = False,
            name: Union[str, None] = None,
            title: str = 'unnamed',
            title_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
            title_anchor: Union[Anchor, str] = Anchor.center, 
            title_style: int = 0,
            *args, **kwargs
        ):
        '''
            title_style: `int` = 0 | 1
        '''
        
        super().__init__(size, offset_pos, widgets, clear_pixel, frame, name, *args, **kwargs)
                
        self._title = title
        self._title_pixel = Converter.toPixel(title_pixel, Pixel((0, 0, 0), (255, 255, 255)))
        self._title_anchor = Converter.toAnchor(title_anchor)
        self._title_buffer: Buffer[Pixel] = Buffer.empty
        self._title_style = title_style
        
        self.resize_event.add(self.setRenderTitle)
        
        self._flag_renderTitle = True

    def setRenderTitle(self):
        self._flag_renderTitle = True
    
    def renderTitle(self):
        match self._title_style:
            case 1:
                self._title_buffer = Drawer.frame(
                    self.width, 3, 
                    self._clear_pixel.front_color, 
                    self._clear_pixel.back_color
                )
                title_tex_length = max(self.width - 2, 0)
                self._title_buffer.paste(
                    1, 1,
                    Buffer(
                        title_tex_length, 1,
                        self._title_pixel,
                        [
                            Pixel.changePixel(self._title_pixel, part) for part in Func.setTextAnchor(self._title, self._title_anchor, title_tex_length)[:title_tex_length]
                        ]
                    )
                )
            
            case _:
                self._title_buffer = Buffer(
                    self.width, 1, 
                    self._title_pixel, 
                    [
                        Pixel.changePixel(self._title_pixel, part) for part in Func.setTextAnchor(self._title, self._title_anchor, self.width)[:self.width]
                    ]
                )
        
        self._offset_pos_widgets = Vec3(
            0, 
            max(self._title_buffer.height-1, 0), 
            0
        )
        self.setUpdateAutoSize()
        
        self._flag_renderTitle = False

    def render(self):
        self._offset_pos_widgets = Vec3(0, 2 if self._title_style == 1 else 0, 0)
        
        buffer = super().render()
        
        if self._flag_renderTitle:
            self.renderTitle()
        
        buffer.paste(0, 0, self._title_buffer, Drawer.mergeFramePixels)
        
        return buffer
    
    @property
    def title(self) -> str:
        return self._title
    @title.setter
    def title(self, value: str):
        self._title = value
        self.setRenderTitle()



from typing import Union, Callable
from datetime import datetime, date
import json

from .Classes import Anchor, Vec2, Vec3
from .Graphic import Window, Widget

def setTextAnchor(text: str, anchor: Anchor, width: Union[int, None] = None, fillchar: chr = ' '):
    target_width = width if width is not None else len(text)
    match anchor:
        case Anchor.center:
            return text.center(target_width, fillchar)
        case Anchor.right:
            return text.rjust(target_width, fillchar)
        case _:
            return text.ljust(target_width, fillchar)

def readFile(path: str) -> str:
    with open(path, 'r', encoding='utf-8') as file:
        return " ".join(file.readlines())

def saveFile(path: str, data: str) -> str:
    with open(path, 'w', encoding='utf-8') as file:
        file.write(data)

def custom_object_hook(data: dict[str, any]):
    for key, value in data.items():
        if isinstance(value, datetime | date):
            data[key] = value.strftime('%d.%m.%Y %H:%M:%S')
        elif isinstance(value, dict):
            data[key] = custom_object_hook(value)
        elif isinstance(value, list | tuple):
            data[key] = list([custom_object_hook(item) for item in value])
    return data

def readJson(path: str, object_pairs_hook: Callable[[tuple[str, any]], tuple[str, any]] = None) -> dict[str, any]:
    return json.loads(readFile(path), object_pairs_hook=object_pairs_hook)

def saveJson(path: str, data: dict[str, any]):
    saveFile(path, json.dumps(data, ensure_ascii=False, indent=4))


def calculateSizeByItems(data: list[Union[Window, Widget]]) -> tuple[Vec2[int], Vec2[int]]:
    width = height = 0
    for item in data:
        width = max(item.offset_x + item.width, width)
        height = max(item.offset_y + item.height, height)
    
    return Vec2(width, height)


import os, sys

from ..Log import Log
from ..Threads import BaseThread
from ..Managers import WindowManager
from ..Graphic import Widget, Window, Pixel, Pixels
from ..GVars import *

class GraphicThread(BaseThread):
    def __init__(self):
        super().__init__(1/GVars.FPS)
    
    async def simulation(self):
        buffer = WindowManager.render()
        if buffer is None:
            return
        
        buffer_data = [pixel if pixel is not None else Pixel.empty for pixel in buffer.data]
        
        pixels: list[Pixel] = \
        [
            buffer_data[i].ANSII if i - i // buffer.width * buffer.width == 0 or not Pixel.compareColors(buffer_data[i-1], buffer_data[i]) else buffer_data[i].symbol 
            #buffer_data[i].symbol
            #buffer_data[i].ANSII
                for i in range(len(buffer))
        ]
        
        rendered_text = ''.join([
            ''.join(pixels[y*buffer.width : y*buffer.width+buffer.width]) + f'{Pixel.clearANSII}\n' for y in range(buffer.height)
        ])
        
        # os.system('cls')
        sys.stdout.write('\n' * GVars.CLEAR_LINES + rendered_text)
    


from typing import Union, Iterable

from ..Widget import Widget
from ..Pixel import Pixel, Pixels
from ...Classes import Vec2, Vec3, Anchor
from ... import Func
from ... import Converter

class Label(Widget):
    def __init__(
        self, 
        text: str = 'label', 
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        anchor: Union[Anchor, str] = Anchor.left,
        tab_length: int = 4,
        *args, **kwargs
    ):
        super().__init__((0, 0), offset_pos, clear_pixel, name, *args, **kwargs)
        
        self._anchor = Converter.toAnchor(anchor)
        self._min_size = Converter.toVec2(min_size, Vec2(None, None), True)
        self._lines: tuple[str] = ()
        self._tab_length = tab_length
        
        self.text = text
    
    def refresh(self):
        super().refresh()
        for y in range(len(self._lines)):
            for x in range(len(self._lines[y])):
                self._buffer[x, y] = Pixel.changePixel(
                    self._clear_pixel if self._clear_pixel is not None else Pixels.f_white, 
                    symbol=self._lines[y][x]
                )
    
    @property
    def text(self) -> str:
        return '\n'.join(self._lines)
    @text.setter
    def text(self, value: str):
        text = str(value).replace('\t', ' '*self._tab_length)
        self._lines = tuple([
            Func.setTextAnchor(
                line, 
                self._anchor, 
                max(len(line), self._min_size.x),
                self._clear_pixel.symbol
            ) if self._min_size.x is not None else line for line in [
                line.replace('\n', '') for line in text.rsplit('\n')
            ]
        ])
        
        self.size = Vec2(
            max(map(len, self._lines)), 
            len(self._lines)
        )
        
        self.setRefresh()




from random import randint as rd
from typing import Union, Iterable, TypeVar

from ..Classes import Buffer, Vec2, Vec3, Event, Counter
from ..Graphic import Pixel

from .. import Converter

class Widget:
    _widgets: dict[str, 'Widget'] = {}
    _counter: Counter = Counter()
    
    @classmethod
    def getByName(cls, name: str) -> Union['Widget', None]:
        return cls._widgets.get(name)
    
    @classmethod
    def tryGetByName(cls, name: str) -> tuple[bool, Union['Widget', None]]:
        widget = cls.getByName(name)
        return (
            widget is not None,
            widget
        )
    
    @classmethod
    def removeAll(cls):
        cls._widgets.clear()
    
    @classmethod
    def generateNewWidgetWithName(cls, widget_class: 'Widget', *args, **kwargs) -> 'Widget':
        if not issubclass(widget_class, cls):
            raise ValueError(f'Class "{widget_class.__name__}" is not subclass {cls.__name__}')
        
        class_name = widget_class.__name__
        cls._counter.add(class_name)
        
        kwargs.update({
            "name": f'{class_name}_{cls._counter.get(class_name)}'
        })
        
        return widget_class(*args, **kwargs)
    
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None, 
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        *args, **kwargs
    ):
        ''' events '''
        self._set_refreshing_event: Event = Event()
        self._resize_event: Event = Event()
        
        ''' size and pos '''
        self.size = Converter.toVec2(size)
        self._offset_pos = Converter.toVec3(offset_pos)
        
        ''' pixels '''
        self._clear_pixel: Union[Pixel, None] = Converter.toPixel(clear_pixel)
        
        ''' buffers '''
        self._buffer: Buffer[Pixel] = None
        
        ''' flags '''
        self._flag_refresh: bool = True
        
        ''' others '''
        self._name = name
        if self._name is not None:
            if self._name in self.__class__._widgets:
                raise ValueError(f'Widget name "{self._name}" already used')
            self.__class__._widgets[self._name] = self
    
    def setRefresh(self):
        self._flag_refresh = True
        self._set_refreshing_event.invoke()
    
    def refresh(self):
        self._buffer = Buffer(*self._size, self._clear_pixel)
        self._flag_refresh = False
    
    def render(self) -> Buffer[Pixel]:
        if self._flag_refresh:
            self.refresh()
            
        return self._buffer if self._buffer is not None else Buffer.empty

    @property
    def offset_pos(self) -> Vec3[int]:
        return self._offset_pos
    @property
    def offset_x(self) -> int:
        return self.offset_pos.x
    @property
    def offset_y(self) -> int:
        return self.offset_pos.y
    @property
    def offset_z(self) -> int:
        return self.offset_pos.z
    
    @property
    def size(self) -> Vec2:
        return self._size
    @size.setter
    def size(self, value: Vec2):
        self._size = Converter.toVec2(value)
        self._size.change_event.add(self.setRefresh)
        self._resize_event.invoke()
        self.setRefresh()
    @property
    def width(self) -> int:
        return self.size.x
    @width.setter
    def width(self, value: int):
        self.size.x = value
    @property
    def height(self) -> int:
        return self.size.y
    @height.setter
    def height(self, value: int):
        self.size.y = value
        
    @property
    def set_refresh_event(self) -> Event:
        return self._set_refreshing_event
    @property
    def resize_event(self) -> Event:
        return self._resize_event
    
    def __str__(self, **kwargs):
        kwargs.update({
            "name": self._name,
            "size": self._size,
            "offset_pos": self._offset_pos
        })
        return f'{self.__class__.__name__}({'; '.join([f'{key}:{value}' for key, value in kwargs.items()])})'


from typing import Union, Iterable

from ...Classes import Vec2, Vec3, Buffer
from .Container import Container
from ..Widget import Widget
from ..Pixel import Pixel, Pixels
from ... import Converter

class Scroll(Widget):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None, 
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = None,
        scroll_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        *args, **kwargs
    ):
        super().__init__(size, offset_pos, clear_pixel, name, *args, **kwargs)

        self._widgets: list[Widget] = Converter.toListWidgets(widgets)

        self._full_buffer: Union[Buffer[Pixel], None] = None
        self._updateFullBuffer()
        self.scroll = Vec2(0, 0)
        self._scroll_pixel = Converter.toPixel(scroll_pixel, Pixel((0, 0, 0), (255, 255, 255)))
        
    def _updateFullBuffer(self):
        max_width = max_height = 0
        
        for widget in self._widgets:
            max_width = max(max_width, widget.size.x + widget.offset_pos.x)
            max_height = max(max_height, widget.size.y + widget.offset_pos.y)

        self._full_buffer = Buffer(max_width, max_height, None)

        for widget in self._widgets:
            self._full_buffer.paste(
                *(widget.offset_pos.toTuple()[:2]),
                widget.render()
            )
    
    def refresh(self):
        if self._full_buffer is None:
            self._updateFullBuffer()
        
        super().refresh()
        
        self._buffer.paste(
            -self._scroll.x, 
            -self._scroll.y, 
            self._full_buffer
        )
        
        hor = self._full_buffer.width > self._buffer.width
        ver = self._full_buffer.height > self._buffer.height
        both = hor and ver
        
        if hor:
            for x in range(self._buffer.width):
                self._buffer[
                    x, 
                    self._buffer.height-1
                ] = self._scroll_pixel
                
            if self._scroll.x > 0:
                self._buffer[
                    0,
                    self._buffer.height-1
                ] = Pixel.changePixel(self._scroll_pixel, symbol='◄')
            
            if self._full_buffer.width - (self._scroll.x + self._buffer.width) > 0:
                self._buffer[
                    self._buffer.width-1 - (1 if both else 0), 
                    self._buffer.height-1
                ] = Pixel.changePixel(self._scroll_pixel, symbol='►')
        
        if ver:
            for y in range(self._buffer.height):
                self._buffer[
                    self._buffer.width-1, 
                    y
                ] = self._scroll_pixel
                
            if self._scroll.y > 0:
                self._buffer[
                    self._buffer.width-1, 
                    0
                ] = Pixel.changePixel(self._scroll_pixel, symbol='▲')
                
            if self._full_buffer.height - (self._scroll.y + self._buffer.height) > 0:
                self._buffer[
                    self._buffer.width-1, 
                    self._buffer.height-1 - (1 if both else 0)
                ] = Pixel.changePixel(self._scroll_pixel, symbol='▼')
        
        if hor and ver:
            self._buffer[self._buffer.width-1, self._buffer.height-1] = Pixel.changePixel(self._scroll_pixel, symbol='/')
                
    @property
    def scroll(self) -> Vec2[int]:
        return self._scroll
    @scroll.setter
    def scroll(self, value: Vec2[int]):
        def _valide_scroll():
            self._scroll._x = min(max(self._scroll.x, 0), max(self._full_buffer.width - self.size.x, 0))
            self._scroll._y = min(max(self._scroll.y, 0), max(self._full_buffer.height - self.size.y, 0))
            self.setRefresh()
        
        self._scroll = value
        self._scroll.change_event.add(_valide_scroll)
        self.setRefresh()
    
        


        
        
        
        
        


from typing import Union, Iterable

from ..Widget import Widget
from ..Pixel import Pixel
from ...Classes import Vec2, Vec3
from ..Drawer import Drawer
from ... import Converter
from ... import Func

class Container(Widget):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None, 
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = None, 
        *args, **kwargs
    ):
        super().__init__(size, offset_pos, clear_pixel, name, *args, **kwargs)
        
        self._auto_size = auto_size
        self._indent_size = Vec2(0, 0)
        
        self._offset_pos_widgets: Vec3[int] = Vec3(0, 0, 0)
        self._widgets: list[Widget] = []
        for widget in Converter.toListWidgets(widgets):
            self.addWidget(widget)
                
        self._flag_update_auto_size = self._auto_size

    def addWidget(self, widget: Widget):
        self._widgets.append(widget)
        widget.set_refresh_event.add(self.setRefresh)
        if self._auto_size:
            widget.resize_event.add(self.setAutoSize)
    
    def setAutoSize(self):
        self._flag_update_auto_size = True
        self.setRefresh()
    
    def _updateAutoSize(self):
        if self._auto_size is False or len(self._widgets) == 0:
            return
        self.size = Func.calculateSizeByItems(self._widgets) + self._indent_size
        
        self._flag_update_auto_size = False
        
    def refresh(self):
        widgets: list[tuple[str, any]] = []
        for widget in sorted(self._widgets, key=lambda widget: widget.offset_pos.z):
            widgets.append((
                (widget.offset_pos + self._offset_pos_widgets).toTuple()[:2], 
                widget.render()
            ))
        
        if self._flag_update_auto_size:
            self._updateAutoSize()
        
        super().refresh()
        for widget in widgets:
            self._buffer.paste(
                *widget[0],
                widget[1]
            )



from typing import Union, Iterable, Callable

from .Classes import Vec2, Vec3, Anchor, Orientation
from .Graphic import Pixel, Pixels, Widget


def toVec2(data: Union[Vec2, Iterable], default: Vec2 = Vec2(0, 0), allow_none: bool = False) -> Vec2:
    if data is None:
        return default
    if allow_none is False and None in data:
        raise ValueError()
    return data if isinstance(data, Vec2) else Vec2(*data)

def toVec3(data: Union[Vec3, Iterable], default: Vec3 = Vec3(0, 0, 0), allow_none: bool = False) -> Vec3:
    if data is None:
        return default
    if allow_none is False and None in data:
        raise ValueError()
    return data if isinstance(data, Vec3) else Vec3(*data)

def toPixel(data: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None], default: Pixel = None) -> Pixel:
    '''
        `data` can be of any `Iterable-Type`
    '''
    if data is None:
        return default
    elif isinstance(data, str):
        return Pixels.__dict__[data]
    elif isinstance(data, Pixel | Iterable):
        return data if isinstance(data, Pixel) else Pixel(*data)

    raise ValueError(f'data({data}) is not Pixel | Iterable')

def toListWidgets(data: Union[Iterable[Widget], Widget, None]) -> list[Widget]:
    if data is None:
        return []
    if isinstance(data, type(Widget)):
        return [data]
    if not isinstance(data, Iterable):
        raise ValueError()
    return [*data]

def toAnchor(anchor: Union[Anchor, str]) -> Anchor:
    if isinstance(anchor, Anchor):
        return anchor
    elif isinstance(anchor, str):
        return Anchor[anchor]
    raise ValueError()

def toOrientation(orientation: Union[Orientation, str]) -> Orientation:
    if isinstance(orientation, Orientation):
        return orientation
    elif isinstance(orientation, str):
        return Orientation[orientation]
    raise ValueError()



import asyncio
import os
from typing import Union
import importlib
import time

from .GVars import GVars
from .Threads import BaseThread, GraphicThread, SimulationThread, InputThread
from .Classes.Event import Event
from .Log import Log
from .Managers import KeyboardManager as KeyM, posix_key, WindowManager
from . import Func

class Core:
    init_event: Event = Event()
    term_event: Event = Event()
    
    init_all_event = BaseThread.init_all_event
    
    GraphicThread = None
    SimulationThread = None
    InputThread = None

    _tasks: list[asyncio.Task] = []
    
    @classmethod
    def start(cls):
        '''
            run async app
        '''
        try:
            for thread in BaseThread._threads.values():
                cls._tasks.append(GVars.ASYNC_EVENT_LOOP.create_task(thread.run(), name=thread.name))
            
            GVars.ASYNC_EVENT_LOOP.run_until_complete(asyncio.wait(cls._tasks))
        
        except KeyboardInterrupt:
            Log.writeWarning('Emergency termination', cls)
            
        except:
            Log.writeError(cls)
            input('Press to continue...')
                
        finally:
            for task in cls._tasks:
                task.cancel()
            GVars.ASYNC_EVENT_LOOP.run_until_complete(asyncio.wait(cls._tasks))
            GVars.ASYNC_EVENT_LOOP.stop()
            
    @classmethod
    def init(cls):
        cls.GraphicThread = GraphicThread()
        cls.SimulationThread = SimulationThread()
        cls.InputThread = InputThread()
        
        KeyM.registerEvent('_close', posix_key.CTRL_C)
        KeyM.bindEvent('_close', '\x00k') # alt+f4
        KeyM.bind('_close', BaseThread.stopAll)
        
        cls.init_event.invoke()
        Log.writeOk('Initialized', cls)
    
    @classmethod
    def term(cls):
        cls.init_event.invoke()
        Log.writeOk('Terminated', cls)
    
    _dynamic_modules: dict[str, dict[str, any]] = {}
    @classmethod
    def addDynamicModule(cls, path: str, name: str):
        if path in cls._dynamic_modules:
            raise ValueError(f'Module "{path}" already exists')
        
        if not os.path.exists(path):
            raise ValueError(f'File "{path}" not exists')
        
        if GVars.WRITE_WARNING_DYNAMIC_MODULE and len(cls._dynamic_modules) == 0:
            Log.writeWarning(
                '\nThis tool is unstable due to its features and may lead to errors\nIt is strongly recommended to only change variables inside the module\nNo complex logic', cls
            )
            time.sleep(1)
            
        
        cls._dynamic_modules[path] = {
            'mtime': os.path.getmtime(path),
            'name': name,
            'module': importlib.import_module(name)
        }
    
    @classmethod
    def checkDynamicModules(cls):
        for path, data in cls._dynamic_modules.items():
            if not os.path.exists(path):
                continue
            
            os.path.dirname
            
            last_mtime = os.path.getmtime(path)
            if data['mtime'] == last_mtime:
                continue
            
            data['module'] = importlib.reload(data['module'])
            data['mtime'] = last_mtime
            
            Log.writeOk(f'module "{path}" updated')
            
    

             


from typing import Union, Iterable

from .Container import Container
from ..Widget import Widget
from ..Pixel import Pixel
from ...Classes import Vec2, Vec3
from ..Drawer import Drawer

class Frame(Container):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None, 
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = None, 
        *args, **kwargs
    ):
        super().__init__(size, auto_size, offset_pos, clear_pixel, name, widgets, *args, **kwargs)
        self._indent_size = Vec2(2, 2)
        self._offset_pos_widgets = Vec3(1, 1, 0)
        
    def refresh(self):
        super().refresh()
        self._buffer.paste(
            0, 0,
            Drawer.frame(
                *self._size, 
                self._clear_pixel.front_color,
                self._clear_pixel.back_color
            ) if self._clear_pixel is not None else
            Drawer.frame(
                *self._size, 
                None, None
            ),
            Drawer.mergeFramePixels
        )
    
    
        




from ..Widget import Widget
from .Label import Label
from .Frame import Frame
from .Container import Container
from .ProgressBar import ProgressBar
from .Scroll import Scroll
from .InteractWidget import InteractiveWidget